# coding: utf-8
"""Test dtool integration."""
__author__ = 'Johannes Laurin Hoermann'
__copyright__ = 'Copyright 2020, IMTEK Simulation, University of Freiburg'
__email__ = 'johannes.hoermann@imtek.uni-freiburg.de, johannes.laurin@gmail.com'
__date__ = 'Apr 27, 2020'

import glob
import json
import logging
import os
import pytest
import ruamel.yaml as yaml

# needs dtool cli for verification

import dtoolcore

from jlhfw.utils.environ import TemporaryOSEnviron
from jlhfw.utils.logging import _log_nested_dict
from jlhfw.fireworks.user_objects.firetasks.dtool_tasks import (
    CreateDatasetTask, FreezeDatasetTask, CopyDatasetTask, FetchItemTask)

FIXTURE_DIR = os.path.abspath(os.path.dirname(__file__))

DATASET_NAME = 'dataset'


def _read_json(file):
    with open(file, 'r') as stream:
        return json.load(stream)


def _read_yaml(file):
    with open(file, 'r') as stream:
        return yaml.safe_load(stream)


def _inject(source, injection, marker):
    """Inserts values into nested dicts at positions marked by marker."""
    logger = logging.getLogger(__name__)
    if isinstance(marker, dict):
        for k, v in marker.items():
            if isinstance(v, str):  # str in marker marks desired mapping
                # inject here
                logger.debug(
                    "'{}' from injection[{}] overrides '{}' in source[{}]."
                    .format(injection[v], v, source[k], k))
                source[k] = injection[v]
            else:
                # descend
                logger.debug("Descending into sub-tree '{}' of '{}'.".format(
                    source[k], source))
                source[k] = _inject(source[k], injection, marker[k])
    elif isinstance(marker, list):  # source and marker must have same length
        logger.debug("Branching into element wise sub-trees of '{}'.".format(
            source))
        source = [_inject(s, injection, m) for s, m in zip(source, marker)]
    else:  # nothing to be done for this key
        logger.debug("No injection desired at '{}'.".format(source))

    return source


def _compare(source, target, marker):
    """Compare source and target partially, as marked by marker."""
    # ret = True
    logger = logging.getLogger(__name__)
    if isinstance(marker, dict):
        for k, v in marker.items():
            if k not in source:
                logger.error("{} not in source '{}'.".format(k, source))
                return False
            if k not in target:
                logger.error("{} not in target '{}'.".format(k, source))
                return False

            logger.debug("Descending into sub-tree '{}' of '{}'.".format(
                source[k], source))
            # descend
            if not _compare(source[k], target[k], v):
                return False  # one failed comparison suffices

    elif isinstance(marker, list):  # source, target and marker must have same length
        logger.debug("Branching into element wise sub-trees of '{}'.".format(
            source))
        for s, t, m in zip(source, target, marker):
            if not _compare(s, t, m):
                return False  # one failed comparison suffices
    else:  # arrived at leaf, comparison desired?
        if marker is not False:  # yes
            logger.debug("Comparing '{}' == '{}' -> {}.".format(
                source, target, source == target))
            return source == target

    # comparison either not desired or successfull for all elements
    return True


def _compare_frozen_metadata_against_template(
        readme_file, template_file, config_file, to_compare):
    logger = logging.getLogger(__name__)
    frozen_readme = _read_yaml(readme_file)
    logger.debug("Frozen README.yml:")
    _log_nested_dict(logger.debug, frozen_readme)

    reference_readme = _read_yaml(template_file)
    logger.debug("Reference README.yml:")
    _log_nested_dict(logger.debug, reference_readme)

    dtool_config = _read_json(config_file)
    logger.debug("dtool config:")
    _log_nested_dict(logger.debug, dtool_config)

    logger.debug("Comaprison mode:")
    _log_nested_dict(logger.debug, to_compare)

    parsed_reference_readme = _inject(
        reference_readme, dtool_config, to_compare)

    logger.debug("Parsed reference README.yml:")
    _log_nested_dict(logger.debug, parsed_reference_readme)

    return _compare(frozen_readme, parsed_reference_readme, to_compare)


# see https://github.com/jic-dtool/dtool-info/blob/64e021dc06cc6dc6df3d5858fda3e553fa18b91d/dtool_info/dataset.py#L354
def verify(full, dataset_uri):
    """Verify the integrity of a dataset."""
    logger = logging.getLogger(__name__)
    dataset = dtoolcore.DataSet.from_uri(dataset_uri)
    all_okay = True

    # Generate identifiers and sizes quickly without the
    # hash calculation used when calling dataset.generate_manifest().
    generated_sizes = {}
    generated_relpaths = {}
    for handle in dataset._storage_broker.iter_item_handles():
        identifier = dtoolcore.utils.generate_identifier(handle)
        size = dataset._storage_broker.get_size_in_bytes(handle)
        relpath = dataset._storage_broker.get_relpath(handle)
        generated_sizes[identifier] = size
        generated_relpaths[identifier] = relpath

    generated_identifiers = set(generated_sizes.keys())
    manifest_identifiers = set(dataset.identifiers)

    for i in generated_identifiers.difference(manifest_identifiers):
        message = "Unknown item: {} {}".format(
            i,
            generated_relpaths[i]
        )
        logger.warning(message)
        all_okay = False

    for i in manifest_identifiers.difference(generated_identifiers):
        message = "Missing item: {} {}".format(
            i,
            dataset.item_properties(i)["relpath"]
        )
        logger.warning(message)
        all_okay = False

    for i in manifest_identifiers.intersection(generated_identifiers):
        generated_size = generated_sizes[i]
        manifest_size = dataset.item_properties(i)["size_in_bytes"]
        if generated_size != manifest_size:
            message = "Altered item size: {} {}".format(
                i,
                dataset.item_properties(i)["relpath"]
            )
            logger.warning(message)
            all_okay = False

    if full:
        generated_manifest = dataset.generate_manifest()
        for i in manifest_identifiers.intersection(generated_identifiers):
            generated_hash = generated_manifest["items"][i]["hash"]
            manifest_hash = dataset.item_properties(i)["hash"]
            if generated_hash != manifest_hash:
                message = "Altered item hash: {} {}".format(
                    i,
                    dataset.item_properties(i)["relpath"]
                )
                logger.warning(message)
                all_okay = False

    return all_okay


#
# pytest fixtures
#
@pytest.fixture
def default_create_dataset_task_spec(dtool_config, files):
    """Provide default test task_spec for CreateDatasetTask."""
    return {
        'name': DATASET_NAME,
        'dtool_readme_template_path': files["dtool_readme_template_path"],
        'dtool_config': dtool_config,
        'stored_data': True,
    }


@pytest.fixture
def default_freeze_dataset_task_spec(dtool_config):
    """Provide default test task_spec for FreezeDatasetTask."""
    return {
        'dtool_config': dtool_config,
        'stored_data': True,
    }


@pytest.fixture
def default_copy_dataset_task_spec(dtool_config):
    """Provide default test task_spec for CopyDatasetTask."""
    return {
        'source': DATASET_NAME,
        'target': 'smb://test-share',
        'dtool_config': dtool_config,
        'stored_data': True,
    }


@pytest.fixture
def default_fetch_item_task_spec(dtool_config):
    """Provide default test task_spec for CopyDatasetTask."""
    return {
        'source': os.path.join(FIXTURE_DIR, 'immutable_test_dataset'),
        'item_id': 'eb58eb70ebcddf630feeea28834f5256c207edfd',
        'filename': 'fetched_item.txt',
        'dtool_config': dtool_config,
        'stored_data': True,
    }


def _create_dataset(tempdir, files, dtool_config,
                    default_create_dataset_task_spec,
                    default_freeze_dataset_task_spec):
    """Will create dataset with default parameters within current working directory."""
    logger = logging.getLogger(__name__)

    logger.debug("Instantiate CreateDatasetTask with '{}'".format(
        default_create_dataset_task_spec))
    t = CreateDatasetTask(**default_create_dataset_task_spec)
    fw_action = t.run_task({})
    logger.debug("FWAction:")
    _log_nested_dict(logger.debug, fw_action.as_dict())
    uri = fw_action.stored_data['uri']

    logger.debug("Instantiate FreezeDatasetTask with '{}'".format(
        {'uri': uri, **default_freeze_dataset_task_spec}))
    t = FreezeDatasetTask(
        uri=uri, **default_freeze_dataset_task_spec)
    fw_action = t.run_task({})
    logger.debug("FWAction:")
    _log_nested_dict(logger.debug, fw_action.as_dict())
    uri = fw_action.stored_data['uri']

    with TemporaryOSEnviron(_read_json(files['dtool_config_path'])):
        ret = verify(True, uri)

    assert ret

    # compare metadata template and generated README.yml
    # expect filled in values like this:
    #    project: DtoolTasksTest
    #    description: Tests on Fireworks tasks for handling dtool datasets.
    #    owners:
    #      - name: Dtool Tasks Test
    #        email: dtool@jlhfw
    #        username: jotelha
    #    creation_date: 2020-04-27
    #    expiration_date: 2020-04-27
    #    metadata:
    #      mode: trial
    #      step: override this

    # legend:
    # - True compares key and value
    # - False confirms key existance but does not compare value
    # - str looks up value from config gile
    to_compare = {
        'project': True,
        'description': True,
        'owners': [{
            'name': 'DTOOL_USER_FULL_NAME',
            'email': 'DTOOL_USER_EMAIL',
            'username': False,
        }],
        'creation_date': False,
        'expiration_date': False,
        'metadata': {
            'mode': True,
            'step': True,
        }
    }

    compares = _compare_frozen_metadata_against_template(
        os.path.join(DATASET_NAME, "README.yml"),
        files['dtool_readme_template_path'],
        files['dtool_config_path'],
        to_compare
    )
    assert compares

    return uri


#
# basic dtool tasks tests
#
def test_create_dataset_task_run(tempdir, files, dtool_config,
                                 default_create_dataset_task_spec,
                                 default_freeze_dataset_task_spec):
    """Will create dataset with default parameters within current working directory."""
    _create_dataset(tempdir, files, dtool_config,
                    default_create_dataset_task_spec,
                    default_freeze_dataset_task_spec)


def test_static_metadata_override(tempdir, files, dtool_config,
                                  default_create_dataset_task_spec,
                                  default_freeze_dataset_task_spec):
    """Will create dataset with static metadata within current working directory."""
    logger = logging.getLogger(__name__)

    t = CreateDatasetTask(
        **default_create_dataset_task_spec,
        metadata={
            'metadata': {
                'step': 'static metadata test'
            }
        }
    )
    fw_action = t.run_task({})
    logger.debug("FWAction:")
    _log_nested_dict(logger.debug, fw_action.as_dict())
    uri = fw_action.stored_data['uri']

    logger.debug("Instantiate FreezeDatasetTask with '{}'".format(
        {'uri': uri, **default_freeze_dataset_task_spec}))
    t = FreezeDatasetTask(
        uri=uri, **default_freeze_dataset_task_spec)
    fw_action = t.run_task({})
    logger.debug("FWAction:")
    _log_nested_dict(logger.debug, fw_action.as_dict())
    uri = fw_action.stored_data['uri']

    with TemporaryOSEnviron(_read_json(files['dtool_config_path'])):
        ret = verify(True, uri)
    assert ret

    # legend:
    # - True compares key and value
    # - False confirms key existance but does not compare value
    # - str looks up value from config gile
    to_compare = {
        'project': True,
        'description': True,
        'owners': [{
            'name': 'DTOOL_USER_FULL_NAME',
            'email': 'DTOOL_USER_EMAIL',
            'username': False,
        }],
        'creation_date': False,
        'expiration_date': False,
        'metadata': {
            'mode': True,
            'step': True,
        }
    }

    compares = _compare_frozen_metadata_against_template(
        os.path.join(DATASET_NAME, "README.yml"),
        files['dtool_readme_static_metadata_test'],
        files['dtool_config_path'],
        to_compare
    )
    assert compares

    return uri


def test_dynamic_metadata_override(tempdir, files, dtool_config,
                                   default_create_dataset_task_spec,
                                   default_freeze_dataset_task_spec):
    """Will create dataset with static and dynamic metadata within current working directory."""
    logger = logging.getLogger(__name__)

    t = CreateDatasetTask(
        **default_create_dataset_task_spec,
        metadata_key='deeply->deeply->nested',
    )
    fw_action = t.run_task(  # insert some fw_spec into metadata
        {
            'deeply': {
                'deeply': {
                    'nested': {
                        'metadata': {
                            'step': 'dynamic metadata test'
                        }
                    }
                }
            }
        }
    )
    logger.debug("FWAction:")
    _log_nested_dict(logger.debug, fw_action.as_dict())
    uri = fw_action.stored_data['uri']

    logger.debug("Instantiate FreezeDatasetTask with '{}'".format(
        {'uri': uri, **default_freeze_dataset_task_spec}))
    t = FreezeDatasetTask(
        uri=uri, **default_freeze_dataset_task_spec)
    fw_action = t.run_task({})
    logger.debug("FWAction:")
    _log_nested_dict(logger.debug, fw_action.as_dict())
    uri = fw_action.stored_data['uri']

    with TemporaryOSEnviron(_read_json(files['dtool_config_path'])):
        ret = verify(True, uri)
    assert ret

    # legend:
    # - True compares key and value
    # - False confirms key existance but does not compare value
    # - str looks up value from config gile
    to_compare = {
        'project': True,
        'description': True,
        'owners': [{
            'name': 'DTOOL_USER_FULL_NAME',
            'email': 'DTOOL_USER_EMAIL',
            'username': False,
        }],
        'creation_date': False,
        'expiration_date': False,
        'metadata': {
            'mode': True,
            'step': True,
        }
    }

    compares = _compare_frozen_metadata_against_template(
        os.path.join(DATASET_NAME, "README.yml"),
        files['dtool_readme_dynamic_metadata_test'],
        files['dtool_config_path'],
        to_compare
    )
    assert compares

    return uri


def test_static_and_dynamic_metadata_override(tempdir, files, dtool_config,
                                              default_create_dataset_task_spec,
                                              default_freeze_dataset_task_spec):
    """Will create dataset with static metadata within current working directory."""
    logger = logging.getLogger(__name__)

    t = CreateDatasetTask(
        **default_create_dataset_task_spec,
        metadata={
            'metadata': {  # insert some task-level specs into metadata
                'step': 'static and dynamic metadata test',  #
                'static_field': 'set by task-level static metadata field'
            }
        },
        metadata_key='deeply->deeply->nested',
    )
    fw_action = t.run_task(  # insert some fw_spec into metadata
        {
            'deeply': {
                'deeply': {
                    'nested': {
                        'metadata': {
                            'step': 'static field above overrides me',
                            'dynamic_field': 'set by Fireworks fw_spec-level dynamic metadata field'
                        }
                    }
                }
            }
        }
    )
    logger.debug("FWAction:")
    _log_nested_dict(logger.debug, fw_action.as_dict())
    uri = fw_action.stored_data['uri']

    logger.debug("Instantiate FreezeDatasetTask with '{}'".format(
        {'uri': uri, **default_freeze_dataset_task_spec}))
    t = FreezeDatasetTask(
        uri=uri, **default_freeze_dataset_task_spec)
    fw_action = t.run_task({})
    logger.debug("FWAction:")
    _log_nested_dict(logger.debug, fw_action.as_dict())
    uri = fw_action.stored_data['uri']

    with TemporaryOSEnviron(_read_json(files['dtool_config_path'])):
        ret = verify(True, uri)
    assert ret

    # legend:
    # - True compares key and value
    # - False confirms key existance but does not compare value
    # - str looks up value from config gile
    to_compare = {
        'project': True,
        'description': True,
        'owners': [{
            'name': 'DTOOL_USER_FULL_NAME',
            'email': 'DTOOL_USER_EMAIL',
            'username': False,
        }],
        'creation_date': False,
        'expiration_date': False,
        'metadata': {
            'mode': True,
            'step': True,
            'static_field': True,
            'dynamic_field': True,
        }
    }

    compares = _compare_frozen_metadata_against_template(
        os.path.join(DATASET_NAME, "README.yml"),
        files['dtool_readme_static_and_dynamic_metadata_test'],
        files['dtool_config_path'],
        to_compare
    )
    assert compares

    return uri


def test_params_from_fw_spec(tempdir, files, dtool_config,
                             default_create_dataset_task_spec,
                             default_freeze_dataset_task_spec):
    """Will create dataset with some task parameters pulled from fw_spec."""
    logger = logging.getLogger(__name__)

    t_spec = {
        **default_create_dataset_task_spec,
        'creator_username': {'key': 'deeply->deeply->nested->username'}
    }
    t = CreateDatasetTask(**t_spec)

    fw_action = t.run_task(
        {
            'deeply': {
                'deeply': {
                    'nested': {
                        'username': 'unittest'
                    }
                }
            }
        }
    )
    logger.debug("FWAction:")
    _log_nested_dict(logger.debug, fw_action.as_dict())
    uri = fw_action.stored_data['uri']

    logger.debug("Instantiate FreezeDatasetTask with '{}'".format(
        {'uri': uri, **default_freeze_dataset_task_spec}))
    t = FreezeDatasetTask(
        uri=uri, **default_freeze_dataset_task_spec)
    fw_action = t.run_task({})
    logger.debug("FWAction:")
    _log_nested_dict(logger.debug, fw_action.as_dict())
    uri = fw_action.stored_data['uri']

    with TemporaryOSEnviron(_read_json(files['dtool_config_path'])):
        ret = verify(True, uri)
    assert ret

    # check creator_username
    dataset = dtoolcore.DataSet.from_uri(uri)
    assert dataset._admin_metadata["creator_username"] == 'unittest'

    return uri


#
# derived datasets tests
#
def test_create_derived_dataset(tempdir, files, dtool_config,
                                default_create_dataset_task_spec,
                                default_freeze_dataset_task_spec):
    """Create two datasets, one derived from the other."""
    logger = logging.getLogger(__name__)

    # create a dummy dataset to derive from
    source_dataset_uri = _create_dataset(tempdir, files, dtool_config,
                                         default_create_dataset_task_spec,
                                         default_freeze_dataset_task_spec)
    logger.debug("Derive from source dataset with URI '{}'.".format(
        source_dataset_uri))

    t_spec = default_create_dataset_task_spec.copy()
    t_spec['name'] = 'derived_datset'
    t_spec['source_dataset_uri'] = source_dataset_uri

    logger.debug("Instantiate another CreateDatasetTask with task spec:")
    _log_nested_dict(logger.debug, t_spec)
    t = CreateDatasetTask(**t_spec)
    fw_action = t.run_task({})
    logger.debug("FWAction:")
    _log_nested_dict(logger.debug, fw_action.as_dict())
    derived_datset_uri = fw_action.stored_data['uri']

    logger.debug("Instantiate FreezeDatasetTask with '{}'".format(
        {'uri': derived_datset_uri, **default_freeze_dataset_task_spec}))
    t = FreezeDatasetTask(
        uri=derived_datset_uri, **default_freeze_dataset_task_spec)
    fw_action = t.run_task({})
    logger.debug("FWAction:")
    _log_nested_dict(logger.debug, fw_action.as_dict())
    derived_dataset_uri = fw_action.stored_data['uri']

    with TemporaryOSEnviron(_read_json(files['dtool_config_path'])):
        ret = verify(True, derived_datset_uri)
    assert ret

    # check that fields for derived datasets agree
    source_dataset = dtoolcore.DataSet.from_uri(source_dataset_uri)
    derived_dataset = dtoolcore.DataSet.from_uri(derived_dataset_uri)
    assert source_dataset.name == derived_dataset.get_annotation("source_dataset_name")
    assert source_dataset.uri == derived_dataset.get_annotation("source_dataset_uri")
    assert source_dataset.uuid == derived_dataset.get_annotation("source_dataset_uuid")


def test_create_derived_dataset_from_multiple_source_dataset_uris(tempdir, files, dtool_config,
                                                                  default_create_dataset_task_spec,
                                                                  default_freeze_dataset_task_spec):
    """Create three datasets, one derived from the other two."""
    logger = logging.getLogger(__name__)

    # create two dummy datasets to derive from

    logger.debug("Instantiate CreateDatasetTask with '{}'".format(
        default_create_dataset_task_spec))

    dummy_names = ['dataset', 'another_dataset']

    # for testing, add a non-existant source uri in the beginning
    source_dataset_ref = [{'uri': 'file://some/none-existant/dataset'}]
    for name in dummy_names:
        t_spec = default_create_dataset_task_spec.copy()
        t_spec['name'] = name
        t = CreateDatasetTask(**t_spec)
        fw_action = t.run_task({})
        logger.debug("FWAction:")
        _log_nested_dict(logger.debug, fw_action.as_dict())
        uri = fw_action.stored_data['uri']
        # source_dataset_uri.append(fw_action.stored_data['uri'])

        logger.debug("Instantiate FreezeDatasetTask with '{}'".format(
            {'uri': uri, **default_freeze_dataset_task_spec}))
        t = FreezeDatasetTask(
            uri=uri, **default_freeze_dataset_task_spec)
        fw_action = t.run_task({})
        logger.debug("FWAction:")
        _log_nested_dict(logger.debug, fw_action.as_dict())
        # uri = fw_action.stored_data['uri']

        with TemporaryOSEnviron(_read_json(files['dtool_config_path'])):
            ret = verify(True, uri)

        assert ret

        to_compare = {
            'project': True,
            'description': True,
            'owners': [{
                'name': 'DTOOL_USER_FULL_NAME',
                'email': 'DTOOL_USER_EMAIL',
                'username': False,
            }],
            'creation_date': False,
            'expiration_date': False,
            'metadata': {
                'mode': True,
                'step': True,
            }
        }

        compares = _compare_frozen_metadata_against_template(
            os.path.join(DATASET_NAME, "README.yml"),
            files['dtool_readme_template_path'],
            files['dtool_config_path'],
            to_compare
        )
        assert compares
        source_dataset_ref.append(
            {
                'name': fw_action.stored_data['name'],
                'uri': fw_action.stored_data['uri'],
                'uuid': fw_action.stored_data['uuid'],

            }
        )

    logger.debug("Derive from source datasets with URIs '{}'.".format(
        source_dataset_ref))

    t_spec = default_create_dataset_task_spec.copy()
    t_spec['name'] = 'derived_datset'
    t_spec['source_dataset_uri'] = [d['uri'] for d in source_dataset_ref]

    logger.debug("Instantiate another CreateDatasetTask with task spec:")
    _log_nested_dict(logger.debug, t_spec)
    t = CreateDatasetTask(**t_spec)
    fw_action = t.run_task({})
    logger.debug("FWAction:")
    _log_nested_dict(logger.debug, fw_action.as_dict())
    derived_datset_uri = fw_action.stored_data['uri']

    logger.debug("Instantiate FreezeDatasetTask with '{}'".format(
        {'uri': derived_datset_uri, **default_freeze_dataset_task_spec}))
    t = FreezeDatasetTask(
        uri=derived_datset_uri, **default_freeze_dataset_task_spec)
    fw_action = t.run_task({})
    logger.debug("FWAction:")
    _log_nested_dict(logger.debug, fw_action.as_dict())
    derived_dataset_uri = fw_action.stored_data['uri']

    with TemporaryOSEnviron(_read_json(files['dtool_config_path'])):
        ret = verify(True, derived_datset_uri)
    assert ret

    # check that fields for derived datasets agree
    # if handled corrctly, then the second uri in the list must have
    # been picked as primary source
    source_dataset = dtoolcore.DataSet.from_uri(source_dataset_ref[1]['uri'])
    derived_dataset = dtoolcore.DataSet.from_uri(derived_dataset_uri)
    assert source_dataset.name == derived_dataset.get_annotation("source_dataset_name")
    assert source_dataset.uri == derived_dataset.get_annotation("source_dataset_uri")
    assert source_dataset.uuid == derived_dataset.get_annotation("source_dataset_uuid")

    # make sure dependeny annotations are set correctly
    source_datasets = derived_dataset.get_annotation("derived_from")
    for ref, cur in zip(source_dataset_ref, source_datasets):
        assert len(ref) == len(cur)
        for k, l in zip(sorted(ref), sorted(cur)):
            assert k == l
            assert ref[k] == cur[l]


def test_create_derived_dataset_from_multiple_source_datasets(tempdir, files, dtool_config,
                                                              default_create_dataset_task_spec,
                                                              default_freeze_dataset_task_spec):
    """Create three datasets, one derived from the other two."""
    logger = logging.getLogger(__name__)

    # create two dummy datasets to derive from

    logger.debug("Instantiate CreateDatasetTask with '{}'".format(
        default_create_dataset_task_spec))

    dummy_names = ['dataset', 'another_dataset']

    # for testing, add a non-existant source uri in the beginning
    source_dataset_ref = [{'uri': 'file://some/none-existant/dataset'}]
    for name in dummy_names:
        t_spec = default_create_dataset_task_spec.copy()
        t_spec['name'] = name
        t = CreateDatasetTask(**t_spec)
        fw_action = t.run_task({})
        logger.debug("FWAction:")
        _log_nested_dict(logger.debug, fw_action.as_dict())
        uri = fw_action.stored_data['uri']

        logger.debug("Instantiate FreezeDatasetTask with '{}'".format(
            {'uri': uri, **default_freeze_dataset_task_spec}))
        t = FreezeDatasetTask(
            uri=uri, **default_freeze_dataset_task_spec)
        fw_action = t.run_task({})
        logger.debug("FWAction:")
        _log_nested_dict(logger.debug, fw_action.as_dict())
        # uri = fw_action.stored_data['uri']

        with TemporaryOSEnviron(_read_json(files['dtool_config_path'])):
            ret = verify(True, uri)

        assert ret

        to_compare = {
            'project': True,
            'description': True,
            'owners': [{
                'name': 'DTOOL_USER_FULL_NAME',
                'email': 'DTOOL_USER_EMAIL',
                'username': False,
            }],
            'creation_date': False,
            'expiration_date': False,
            'metadata': {
                'mode': True,
                'step': True,
            }
        }

        compares = _compare_frozen_metadata_against_template(
            os.path.join(DATASET_NAME, "README.yml"),
            files['dtool_readme_template_path'],
            files['dtool_config_path'],
            to_compare
        )
        assert compares
        source_dataset_ref.append(
            {
                'name': fw_action.stored_data['name'],
                'uri': fw_action.stored_data['uri'],
                'uuid': fw_action.stored_data['uuid'],

            }
        )

    logger.debug("Derive from source datasets with URIs '{}'.".format(
        source_dataset_ref))

    t_spec = default_create_dataset_task_spec.copy()
    t_spec['name'] = 'derived_datset'
    t_spec['source_dataset'] = source_dataset_ref

    logger.debug("Instantiate another CreateDatasetTask with task spec:")
    _log_nested_dict(logger.debug, t_spec)
    t = CreateDatasetTask(**t_spec)
    fw_action = t.run_task({})
    logger.debug("FWAction:")
    _log_nested_dict(logger.debug, fw_action.as_dict())
    derived_datset_uri = fw_action.stored_data['uri']

    logger.debug("Instantiate FreezeDatasetTask with '{}'".format(
        {'uri': derived_datset_uri, **default_freeze_dataset_task_spec}))
    t = FreezeDatasetTask(
        uri=derived_datset_uri, **default_freeze_dataset_task_spec)
    fw_action = t.run_task({})
    logger.debug("FWAction:")
    _log_nested_dict(logger.debug, fw_action.as_dict())
    derived_dataset_uri = fw_action.stored_data['uri']

    with TemporaryOSEnviron(_read_json(files['dtool_config_path'])):
        ret = verify(True, derived_datset_uri)
    assert ret

    # check that fields for derived datasets agree
    # if handled corrctly, then the second uri in the list must have
    # been picked as primary source
    source_dataset = dtoolcore.DataSet.from_uri(source_dataset_ref[1]['uri'])
    derived_dataset = dtoolcore.DataSet.from_uri(derived_dataset_uri)
    assert source_dataset.name == derived_dataset.get_annotation("source_dataset_name")
    assert source_dataset.uri == derived_dataset.get_annotation("source_dataset_uri")
    assert source_dataset.uuid == derived_dataset.get_annotation("source_dataset_uuid")

    # make sure dependeny annotations are set correctly
    source_datasets = derived_dataset.get_annotation("derived_from")
    for ref, cur in zip(source_dataset_ref, source_datasets):
        assert len(ref) == len(cur)
        for k, l in zip(sorted(ref), sorted(cur)):
            assert k == l
            assert ref[k] == cur[l]


#
# copy datasets tests
#
# @pytest.mark.usefixtures("smb_share")
def test_copy_dataset_task_to_smb_share(tempdir, files, dtool_config,
                                        default_create_dataset_task_spec,
                                        default_freeze_dataset_task_spec,
                                        default_copy_dataset_task_spec,
                                        smb_share):
    """Requires a guest-writable share named 'sambashare' available locally.

    The share has to offer an empty sub-directory 'dtool'.

    A snippet within /ect/samba/smb.conf for tetsing purposes may look like
    this:

        [sambashare]
            comment = Samba on Ubuntu
            path = /tmp/sambashare
            guest ok = yes
            available = yes
            read only = no
            writeable = yes
            browsable = yes
            create mask = 0664
            directory mask = 0775
            guest account = unix-user-with-write-privilieges
            force user = unix-user-with-write-privilieges
            force group = sambashare

    You may as well modify access parameters within 'dtool.json' config.
    """
    logger = logging.getLogger(__name__)

    if smb_share is None:
        pytest.skip("No smb share provided.")

    username = smb_share.username
    password = smb_share.password
    host_name = smb_share.remote_name
    host_ip, port = smb_share.sock.getpeername()

    smb_share.createDirectory("sambashare", "dtool")

    # create a dummy dataset locally for transferring to share
    _create_dataset(tempdir, files, dtool_config,
                    default_create_dataset_task_spec,
                    default_freeze_dataset_task_spec)

    dynamic_dtool_config_overrides = {
         "DTOOL_SMB_SERVER_NAME_test-share": host_name,
         "DTOOL_SMB_SERVER_PORT_test-share": port,
         "DTOOL_SMB_USERNAME_test-share": username,
         "DTOOL_SMB_PASSWORD_test-share": password,
         "DTOOL_SMB_DOMAIN_test-share": "WORKGROUP",  # TODO extract
         "DTOOL_SMB_SERVICE_NAME_test-share": "sambashare",
         "DTOOL_SMB_PATH_test-share": "dtool"
    }
    dtool_config.update(dynamic_dtool_config_overrides)
    default_copy_dataset_task_spec['dtool_config'].update(dynamic_dtool_config_overrides)

    logger.debug("dtool_config smb overrides:")
    _log_nested_dict(logger.debug, dtool_config)

    logger.debug("default_copy_dataset_task_spec smb overrides:")
    _log_nested_dict(logger.debug, default_copy_dataset_task_spec)

    logger.debug("Instantiate CopyDatasetTask.")
    t = CopyDatasetTask(**default_copy_dataset_task_spec)

    fw_action = t.run_task({})
    logger.debug("FWAction:")
    _log_nested_dict(logger.debug, fw_action.as_dict())
    target_uri = fw_action.stored_data['uri']

    with TemporaryOSEnviron(dtool_config):
        ret = verify(True, target_uri)
    assert ret

    return target_uri


def test_fetch_item_task(tempdir, files, dtool_config,
                         default_fetch_item_task_spec):
    logger = logging.getLogger(__name__)

    # create a dummy dataset locally for extracting item
    # _create_dataset(tempdir, files, dtool_config,
    #                default_create_dataset_task_spec,
    #                default_freeze_dataset_task_spec)


    logger.debug("Instantiate FetchItemTask.")
    t = FetchItemTask(**default_fetch_item_task_spec)

    fw_action = t.run_task({})
    logger.debug("FWAction:")
    _log_nested_dict(logger.debug, fw_action.as_dict())
    source_uri = fw_action.stored_data['uri']
    with TemporaryOSEnviron(dtool_config):
        ret = verify(True, source_uri)
    assert ret

    local_item_files = glob.glob('*.txt')
    assert len(local_item_files)  == 1
    assert local_item_files[0] == 'fetched_item.txt'
    with open('fetched_item.txt', 'r') as f:
        content = f.read ()
    assert content == 'Some test content'

    return source_uri


# coding: utf-8
#
# dtool_tasks.py
#
# Copyright (C) 2020 IMTEK Simulation
# Author: Johannes Hoermann, johannes.hoermann@imtek.uni-freiburg.de
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Simple Fireworks interface for handling dtool datasets."""

from abc import abstractmethod
from contextlib import ExitStack

import copy  # for deep copies of nested objects, avoid references in YAML dumps
import datetime
import getpass  # get system username for dtool metadata
import glob
import io
import logging
import os
import shutil
import uuid

from ruamel.yaml import YAML

# from dtoolcore import DataSetCreator
import dtoolcore
import dtoolcore.utils
import dtool_create.dataset

from fireworks.fw_config import FW_LOGGING_FORMAT

from fireworks.core.firework import FWAction
from fireworks.utilities.dict_mods import get_nested_dict_value
from fireworks.utilities.fw_serializers import ENCODING_PARAMS

from jlhfw.utils.environ import TemporaryOSEnviron
from jlhfw.utils.dict import simple_dict_merge as dict_merge
from jlhfw.utils.logging import LoggingContext, _log_nested_dict
from jlhfw.utils.multiprocessing import RunAsChildProcessTask


__author__ = 'Johannes Laurin Hoermann'
__copyright__ = 'Copyright 2020, IMTEK Simulation, University of Freiburg'
__email__ = 'johannes.hoermann@imtek.uni-freiburg.de, johannes.laurin@gmail.com'
__date__ = 'Apr 27, 2020'

DEFAULT_FORMATTER = logging.Formatter(FW_LOGGING_FORMAT)


def from_fw_spec(param, fw_spec):
    """Expands param['key'] as key within fw_spec.

    If param is dict hand has field 'key', then return value at specified
    position from fw_spec. Otherwise, return 'param' itself.
    """
    if isinstance(param, dict) and 'key' in param:
        ret = get_nested_dict_value(fw_spec, param['key'])
    else:
        ret = param
    return ret


# adapted from https://github.com/jic-dtool/dtool-create/blob/0a772aa5157523a7219963803293d4e521bc1aa2/dtool_create/dataset.py#L40
def _get_readme_template(fpath=None):
    if fpath is None:
        fpath = dtoolcore.utils.get_config_value(
            "DTOOL_README_TEMPLATE_FPATH",
        )
    if fpath is None:
        fpath = dtool_create.dataset.README_TEMPLATE_FPATH

    with open(fpath, mode='r', **ENCODING_PARAMS) as fh:
        readme_template = fh.read()

    user_email = dtoolcore.utils.get_config_value(
        "DTOOL_USER_EMAIL",
        default="you@example.com"
    )

    user_full_name = dtoolcore.utils.get_config_value(
        "DTOOL_USER_FULL_NAME",
        default="Your Name"
    )

    readme_template = readme_template.format(
        username=getpass.getuser(),
        DTOOL_USER_FULL_NAME=user_full_name,
        DTOOL_USER_EMAIL=user_email,
        date=datetime.date.today(),
    )

    return readme_template


class DtoolTask(RunAsChildProcessTask):
    """
    A dtool task ABC.

    Required params:
        None
    Optional params:
        - dtool_config (dict): dtool config key-value pairs, override
            defaults in $HOME/.config/dtool/dtool.json. Default: None
        - dtool_config_key (str): key to dict within fw_spec, override
            defaults in $HOME/.config/dtool/dtool.json and static dtool_config
            task spec. Default: None.
        - output (str): spec key that will be used to pass
            the handled dataset's properties to child fireworks. Default: None
        - dict_mod (str, default: '_set'): how to insert handled dataset's
            properties into output key, see fireworks.utils.dict_mods
        - propagate (bool, default:None): if True, then set the
            FWAction 'propagate' flag and propagate updated fw_spec not only to
            direct children, but to all descendants down to wokflow's leaves.
        - stored_data (bool, default: False): put handled dataset properties
            into FWAction.stored_data
        - store_stdlog (bool, default: False): insert log output into database
        - stdlog_file (str, Default: NameOfTaskClass.log): print log to file
        - loglevel (str, Default: logging.INFO): loglevel for this task
    """
    _fw_name = 'DtoolTask'
    required_params = [*RunAsChildProcessTask.required_params]
    optional_params = [
        *RunAsChildProcessTask.optional_params,
        "dtool_config",
        "dtool_config_key",
        "stored_data",
        "output",
        "dict_mod",
        "propagate",
        "stdlog_file",
        "store_stdlog",
        "loglevel"]

    @abstractmethod
    def _run_task_internal(self, fw_spec) -> dtoolcore.DataSet:
        """Derivatives implement their functionality here."""
        ...

    def _run_task_as_child_process(self, fw_spec, q, e=None):
        """q is a Queue used to return fw_action."""
        stored_data = self.get('stored_data', False)
        output_key = self.get('output', None)
        dict_mod = self.get('dict_mod', '_set')
        propagate = self.get('propagate', False)

        dtool_config = self.get("dtool_config", {})
        dtool_config_key = self.get("dtool_config_key")

        stdlog_file = self.get('stdlog_file', '{}.log'.format(self._fw_name))
        store_stdlog = self.get('store_stdlog', False)

        loglevel = self.get('loglevel', logging.INFO)

        with ExitStack() as stack:

            if store_stdlog:
                stdlog_stream = io.StringIO()
                logh = logging.StreamHandler(stdlog_stream)
                logh.setFormatter(DEFAULT_FORMATTER)
                stack.enter_context(
                    LoggingContext(handler=logh, level=loglevel, close=False))

            # logging to dedicated log file if desired
            if stdlog_file:
                logfh = logging.FileHandler(stdlog_file, mode='a', **ENCODING_PARAMS)
                logfh.setFormatter(DEFAULT_FORMATTER)
                stack.enter_context(
                    LoggingContext(handler=logfh, level=loglevel, close=True))

            logger = logging.getLogger(__name__)

            logger.debug("task spec level dtool config overrides:")
            _log_nested_dict(logger.debug, dtool_config)

            # fw_spec dynamic dtool_config overrides
            dtool_config_update = {}
            if dtool_config_key is not None:
                try:
                    dtool_config_update = get_nested_dict_value(
                        fw_spec, dtool_config_key)
                    logger.debug("fw_spec level dtool config overrides:")
                    _log_nested_dict(logger.debug, dtool_config_update)
                except Exception:  # key not found
                    logger.warning("{} not found within fw_spec, ignored.".format(
                        dtool_config_key))
            dtool_config.update(dtool_config_update)
            logger.debug("effective dtool config overrides:")
            _log_nested_dict(logger.debug, dtool_config)

            stack.enter_context(TemporaryOSEnviron(env=dtool_config))

            dataset = self._run_task_internal(fw_spec)

        output = {
            'uri': dataset.uri,
            'uuid': dataset.uuid,
            'name': dataset.name,
        }

        if store_stdlog:
            stdlog_stream.flush()
            output['stdlog'] = stdlog_stream.getvalue()

        fw_action = FWAction()

        if stored_data:
            fw_action.stored_data = output

        # 'propagate' only development feature for now
        if hasattr(fw_action, 'propagate') and propagate:
            fw_action.propagate = propagate

        if output_key:  # inject into fw_spec
            fw_action.mod_spec = [{dict_mod: {output_key: output}}]

        # return fw_action
        q.put(fw_action)


class CreateDatasetTask(DtoolTask):
    """
    A Firetask to create a dtool data set.

    This task extends the basic DtoolTask parameters by

    Required params:
        None
    Optional params:
        - creator_username (str): Overrides system username if specified.
            Default: None.
        - directory (str): Path to directory where to do pattern matching.
            Per default '.', i.e. current working directory.
        - dtool_readme_template_path (str): Override default dtool readme
            template path. Default: None.
        - metadata_key (str): If not None, then get additional dynamic
            metadata from the specified key within 'specs' and merge with
            README.yml template and static 'metadata' dict.
            Specified key must point to a dict.
            Specify nested fields with a '->' or '.' delimiter.
            If specified key does not exist or is no dict, then ignore.
            Default: 'metadata'.
        - metadata (dict): Static metadata to attach to data set.
            Static metadata takes precendence over dynamic metadata in case of
            overlap. Also overrides fields specified in readme template.
        - name (str): Name of dataset, default is a v1 UUID.
        - paths (list of str / str): Either list of paths or a glob pattern
            string. Per default, all content of 'directory'.
        - source_dataset_uri (list of str / str): A derived dataset will be
            created if  specified. This is only possible if the source dataset
            is accessible. If this is not the case, this task will raise a
            warning and fall back to the standard dataset creation mechanism.
            If multiple source datasets are specified as a list,
            then the dataset will be derived directly from the first list entry
            If the source datasets are accessible, their names ad UUIDs are
            queried and stored along with their URIs in a list of format

                derived_from:  # or other  label set via 'source_dataset_key'
                - uuid: UUID of the first source dataset if available
                  name: name of the first source dataset if available
                  uri:  URI of the first source dataset
                - uuid: UUID of the second source dataset if available
                  name: name of the second source dataset if available
                  uri:  URI of the second source dataset
                - ...

            in the README.yml. UUID and name fields are omitted if datasets
            are not accessible.
        - source_dataset (list of dict [{str:str}] / dict {str:str}): An
            alternative to above's 'source_dataset_uri'. Can directly specify
            all source datasets' names, uris, and uuids in the format

                source_dataset:
                - uuid: UUID of the first source dataset if available
                  name: name of the first source dataset if available
                  uri:  URI of the first source dataset
                - uuid: UUID of the second source dataset if available
                  name: name of the second source dataset if available
                  uri:  URI of the second source dataset
                - ...

            None of the fields is compulsory. If specified datasets
            are not accessible, then warnings will be raised, but specified
            values will be used as defaults for filling references to sources
            as above. If datasets are accessible, but their names and uuids
            won't agree with the specified defaults, warnings are raised
            accordingly and the actual values will override the defaults.
            If both 'source_dataset_uri' and 'source_dataset' are specified,
            then 'source_dataset_uri' is ignored. Default: None.
        - source_dataset_annotation (bool): If True, then also store
            above's 'derived_from' field as JSON annotation. Default: True.
        - source_dataset_key (str): Key of dependency README.yml entry and
            annotation. Default: 'derived_from'.

    The dataset's README.yml file is a successive merge of the README template,
    static metadata and dynamic metadata, ordered by increasing precedence in
    the case of conflicting fields.

    Fields 'creator_username', 'dtool_readme_template_path', 'name',
    'source_dataset_uri', 'source_dataset', and 'source_dataset_annotation'
    may also be a dict of format { 'key': 'some->nested->fw_spec->key' } for
    looking up value within 'fw_spec' instead.
    """
    _fw_name = 'CreateDatasetTask'
    required_params = [
        *DtoolTask.required_params]

    optional_params = [
        *DtoolTask.optional_params,
        "creator_username",
        "directory",
        "dtool_readme_template_path",
        "metadata",
        "metadata_key",
        "name",
        "paths",
        "source_dataset_uri",
        "source_dataset",
        "source_dataset_annotation",
        "source_dataset_key"]

    def _run_task_internal(self, fw_spec):
        logger = logging.getLogger(__name__)

        name = self.get(
            "name", str(uuid.uuid1()))
        name = from_fw_spec(name, fw_spec)

        # see https://github.com/jic-dtool/dtoolcore/blob/6aff99531d1192f86512f662caf22a6ecd2198a5/dtoolcore/utils.py#L254
        if not dtoolcore.utils.name_is_valid(name):
            raise ValueError((
                "The dataset name can only be 80 characters long. "
                "Valid characters: Alpha numeric characters [0-9a-zA-Z]"
                "Valid special characters: - _ ."))
        logger.info("Create dtool dataset '{}'.".format(name))

        creator_username = self.get(
            "creator_username", None)
        creator_username = from_fw_spec(creator_username, fw_spec)
        if creator_username is not None:
            logger.info("Overriding system username with '{}'.".format(
                creator_username))

        source_dataset_uri = self.get(
            "source_dataset_uri", None)
        source_dataset_uri = from_fw_spec(source_dataset_uri, fw_spec)
        # make sure source_dataset_uri is either str or list of str:
        if isinstance(source_dataset_uri, str):
            source_dataset_uri = [source_dataset_uri]
        if source_dataset_uri is not None:
            try:
                iter(source_dataset_uri)
            except TypeError as exc:
                logger.error("'source_dataset_uri' must be str or list of str.")
                raise exc
            logger.info("Derive from 'source_dataset_uri': '{}'.".format(source_dataset_uri))

        source_dataset = self.get(
            "source_dataset", None)
        source_dataset = from_fw_spec(source_dataset, fw_spec)
        # make sure source_dataset_uri is either dict or list of dict:
        if isinstance(source_dataset, dict):
            source_dataset = [source_dataset]
        if source_dataset is not None:
            try:
                iter(source_dataset)
            except TypeError as exc:
                logger.error("'source_dataset' must be dict or list of dict.")
                raise exc
            logger.info("Derive from 'source_dataset': '{}'.".format(source_dataset))

        if (source_dataset_uri is not None) and (source_dataset is not None):
            logger.info(
                "Both 'source_dataset_uri' and 'source_dataset' are set."
                "Latter will override former.")
        elif source_dataset_uri is not None:  # and not source_dataset
            source_dataset = [{'uri': uri} for uri in source_dataset_uri]

        # from here on, only use source_dataset scheme

        # fill missing fields in source_dataset if uri is present and accessible
        if source_dataset is not None:
            for i, d in enumerate(source_dataset):
                if 'uri' in d:
                    # check whether source dataset accessible
                    try:
                        current_source_dataset = dtoolcore.DataSet.from_uri(
                            d['uri'])
                    except dtoolcore.DtoolCoreTypeError as exc:
                        logger.warning(
                            "Source dataset #{} with URI '{}' not accessible [{}]."
                            .format(i, d['uri'], exc))
                    else:
                        logger.info(
                            "Source dataset #{} has name '{}', uri '{}', and uuid '{}'."
                            .format(i,
                                    current_source_dataset.name,
                                    current_source_dataset.uri,
                                    current_source_dataset.uuid))

                        if 'name' in d and d['name'] != current_source_dataset.name:
                            logger.warning((
                                "Source dataset #{} actual name '{}' "
                                "does not agree with specified default '{}'. "
                                "Former will override latter.").format(
                                    i, current_source_dataset.name, d["name"]))
                        d["name"] = current_source_dataset.name

                        if 'uuid' in d and d['uuid'] != current_source_dataset.uuid:
                            logger.warning((
                                "Source dataset #{} actual UUID '{}' "
                                "does not agree with specified default '{}'. "
                                "Former will override latter.").format(
                                    i, current_source_dataset.uuid, d["uuid"]))
                        d["uuid"] = current_source_dataset.uuid
                else:
                    logger.warning("Source dataset '{}' has no URI specified."
                                .format(d))

        source_dataset_annotation = self.get(
            "source_dataset_annotation", True)
        source_dataset_annotation = from_fw_spec(source_dataset_annotation, fw_spec)

        source_dataset_key = self.get(
            "source_dataset_key", "derived_from")
        source_dataset_key = from_fw_spec(source_dataset_key, fw_spec)

        dtool_readme_template_path = self.get(
            "dtool_readme_template_path", None)
        dtool_readme_template_path = from_fw_spec(dtool_readme_template_path, fw_spec)
        logger.info("Use dtool README.yml template '{}'.".format(
            dtool_readme_template_path))

        dtool_readme_template = _get_readme_template(
            dtool_readme_template_path)
        logger.debug("dtool README.yml template content:")
        _log_nested_dict(logger.debug, dtool_readme_template)

        # generate list of files to place tihin dataset
        directory = os.path.abspath(self.get("directory", "."))
        paths = self.get("paths", None)
        abspaths = []
        if paths is not None:
            if isinstance(self["paths"], list):
                logger.info("Treating 'paths' field as list of files.")
                relpaths = paths
                abspaths = [os.path.abspath(p) for p in self["paths"]]
            else:
                logger.info("Treating 'paths' field as glob pattern.")
                logger.info("Searching within '{}'.".format(directory))
                relpaths = list(
                    glob.glob("{}/{}".format(directory, self["paths"])))
                abspaths = [os.path.abspath(p) for p in relpaths]
        else:  # everything within cwd
            logger.info("'paths' not specified,")
            logger.info("Adding all content of '{}'.".format(directory))
            relpaths = []
            abspaths = []
            # just list files, not directories
            for root, subdirs, files in os.walk('.'):
                for f in files:
                    p = os.path.join(root, f)
                    relpaths.append(p)
                    abspaths.append(os.path.abspath(p))

        logger.debug("Items to add to dataset:")
        logger.debug(relpaths)
        logger.debug("Corresponding absolute paths on file system:")
        logger.debug(abspaths)

        # README.yml metadata merging:

        # see https://github.com/jic-dtool/dtool-create/blob/0a772aa5157523a7219963803293d4e521bc1aa2/dtool_create/dataset.py#L244
        yaml = YAML()
        yaml.explicit_start = True
        yaml.indent(mapping=2, sequence=4, offset=2)
        template_metadata = yaml.load(dtool_readme_template)

        logger.debug("Parsed readme template metadata:")
        _log_nested_dict(logger.debug, template_metadata)

        # fw_spec dynamic metadata
        metadata_key = self.get("metadata_key", "metadata")
        try:
            dynamic_metadata = get_nested_dict_value(fw_spec, metadata_key)
        except Exception:  # key not found
            dynamic_metadata = {}

        logger.debug("Dynamic fw_spec metadata:")
        _log_nested_dict(logger.debug, dynamic_metadata)

        # fw_task static metadata
        static_metadata = self.get("metadata", {})
        logger.debug("Static task-level metadata:")
        _log_nested_dict(logger.debug, static_metadata)

        # source_dataset will override anything else in designated dependency
        # field if specified:
        dependency_metadata = {}
        if source_dataset is not None:
            dependency_metadata = {source_dataset_key: copy.deepcopy(source_dataset)}
            logger.debug("Dependency metadata:")
            _log_nested_dict(logger.debug, dependency_metadata)
        # without deepcopy, references in the dumped structure might appear as
        #   derived_from:
        #     - *id001
        #  in the YAML output instead of the desired
        #   derived_from:  # or other  label set via 'source_dataset_key'
        #     - uuid: UUID of the first source dataset if available
        #       name: name of the first source dataset if available
        #       uri:  URI of the first source dataset

        metadata = {}
        metadata = dict_merge(template_metadata, dynamic_metadata)
        metadata = dict_merge(metadata, static_metadata)
        metadata = dict_merge(metadata, dependency_metadata)
        logger.debug("Merged metadata:")
        _log_nested_dict(logger.debug, metadata)

        readme_stream = io.StringIO()
        yaml.dump(metadata, readme_stream)
        readme = readme_stream.getvalue()

        logger.debug("Content of generated README.yml:")
        for l in readme.splitlines():
            logger.debug("  " + l)

        # dtool default owner metadata for dataset creation
        # admin_metadata = dtoolcore.generate_admin_metadata(name)
        # looks like
        # >>> admin_metadata = dtoolcore.generate_admin_metadata("test")
        # >>> admin_metadata
        # {
        #   'uuid': 'c19bab1d-8f6b-4e44-99ad-1b09ea198dfe',
        #   'dtoolcore_version': '3.17.0',
        #   'name': 'test',
        #   'type': 'protodataset',
        #   'creator_username': 'jotelha',
        #   'created_at': 1587987156.127988
        # }
        # logger.debug("Dataset admin metadata:")
        # _log_nested_dict(logger.debug, admin_metadata)

        # see https://github.com/jic-dtool/dtool-create/blob/0a772aa5157523a7219963803293d4e521bc1aa2/dtool_create/dataset.py#L120
        parsed_base_uri = dtoolcore.utils.generous_parse_uri(os.getcwd())
        # create dataset in current working directory

        # NOTE: dtool_create.dataset.create has specific symlink scheme treatment at
        # https://github.com/jic-dtool/dtool-create/blob/0a772aa5157523a7219963803293d4e521bc1aa2/dtool_create/dataset.py#L127
        # not treated here.

        first_source_dataset = None
        if source_dataset is not None:
            logger.info("Select first accessible source dataset.")
            # find first source dataset with valid URI
            for i, d in enumerate(source_dataset):
                if 'uri' in d:
                    # parsed_source_dataset_uri = dtoolcore.utils.generous_parse_uri(d['uri'])
                    # logger.info("Derive new dataset from '{}'.".format(
                    #     parsed_source_dataset_uri))
                    try:
                        first_source_dataset = dtoolcore.DataSet.from_uri(d["uri"])
                    except dtoolcore.DtoolCoreTypeError as exc:
                        logger.warning("Source dataset #{} with URI '{}' not accessible ( ), skipped.".format(
                                       i, d['uri'], exc))
                    else:
                        logger.info("Found source dataset #{} with URI '{}' accessible.".format(i, d['uri']))
                        logger.info(
                            "Source dataset has name '{}', uri '{}', and uuid '{}'."
                            .format(first_source_dataset.name, first_source_dataset.uri, first_source_dataset.uuid))
                        break
                else:
                    logger.warning("Source dataset #{} has no URI field, skipped.".format(i))
            if first_source_dataset is None:
                logger.warning((
                    "None of specified source datasets accessible. "
                    "Falling back to default dataset creator."))

        if first_source_dataset is not None:
            proto_dataset = dtoolcore.create_derived_proto_dataset(
                name=name,
                base_uri=dtoolcore.utils.urlunparse(parsed_base_uri),
                source_dataset=first_source_dataset,
                readme_content=readme,
                creator_username=creator_username)
        else:   # fall back to default mechanism without source dataset instead
            proto_dataset = dtoolcore.create_proto_dataset(
                name=name,
                base_uri=dtoolcore.utils.urlunparse(parsed_base_uri),
                readme_content=readme,
                creator_username=creator_username)

        # store dependencies as annotations if desired
        if source_dataset_annotation and source_dataset is not None:
            proto_dataset.put_annotation(source_dataset_key, source_dataset)

        # add items to dataset one by one
        # TODO: possibility for per-item metadata
        for abspath, relpath in zip(abspaths, relpaths):
            logger.info("Add '{}' as '{}' to dataset '{}'.".format(
                abspath, relpath, proto_dataset.name))
            proto_dataset.put_item(abspath, relpath)

        logger.info(
            "Created new dataset with name '{}', uri '{}', and uuid '{}'."
            .format(proto_dataset.name, proto_dataset.uri, proto_dataset.uuid))

        return proto_dataset


class FreezeDatasetTask(DtoolTask):
    """
    A Firetask to freeze a dtool data set.

    This task extends the basic DtoolTask parameters by

    Required params:
        None
    Optional params:
        - uri (str): URI of dataset, default: "dataset"

    Field 'uri' may also be a dict of format
    { 'key': 'some->nested->fw_spec->key' } for looking up value within
    fw_spec instead.
    """
    _fw_name = 'FreezeDatasetTask'
    required_params = [*DtoolTask.required_params]
    optional_params = [
        *DtoolTask.optional_params,
        "uri"]

    def _run_task_internal(self, fw_spec):
        logger = logging.getLogger(__name__)

        uri = self.get("uri", "dataset")
        uri = from_fw_spec(uri, fw_spec)

        proto_dataset = dtoolcore.ProtoDataSet.from_uri(uri)
        logger.info("Freezing dataset '{}' with URI '{}'.".format(
            proto_dataset.name, proto_dataset.uri))

        # freeze dataset
        # see https://github.com/jic-dtool/dtool-create/blob/0a772aa5157523a7219963803293d4e521bc1aa2/dtool_create/dataset.py#L438

        num_items = len(list(proto_dataset._identifiers()))
        logger.info("{} items in dataset '{}'.".format(num_items, proto_dataset.name))
        max_files_limit = int(dtoolcore.utils.get_config_value(
            "DTOOL_MAX_FILES_LIMIT",
            default=10000
        ))

        assert isinstance(max_files_limit, int)
        if num_items > max_files_limit:
            raise ValueError(
                "Too many items ({} > {}) in proto dataset".format(
                    num_items,
                    max_files_limit
                ))

        handles = [
            h for h in proto_dataset._storage_broker.iter_item_handles()]
        for h in handles:
            if not dtool_create.utils.valid_handle(h):
                raise ValueError("Invalid item name: {}".format(h))

        logger.debug("Item handles:")
        _log_nested_dict(logger.debug, handles)
        proto_dataset.freeze()

        return proto_dataset


# see https://github.com/jic-dtool/dtool-create/blob/0a772aa5157523a7219963803293d4e521bc1aa2/dtool_create/dataset.py#L494
class CopyDatasetTask(DtoolTask):
    """Copy a dtool data set from source to target.

    This task extends the basic DtoolTask parameters by

    Required params:
        - target (str): base URI of target
    Optional params:
        - source (str): URI of source dataset, default: "dataset".
        - resume (bool): continue to copy a dataset existing already partially
            at target.

    Fields 'target', 'source', and 'resume' may also be a dict of format
    { 'key': 'some->nested->fw_spec->key' } for looking up value within
    fw_spec instead.
    """

    _fw_name = 'CopyDatasetTask'
    required_params = [
        *DtoolTask.required_params,
        "target"]
    optional_params = [
        *DtoolTask.optional_params,
        "source",
        "resume"]

    def _run_task_internal(self, fw_spec):
        logger = logging.getLogger(__name__)

        source = self.get("source", "dataset")
        source = from_fw_spec(source, fw_spec)

        target = self.get("target")
        target = from_fw_spec(target, fw_spec)

        resume = self.get("resume", False)
        resume = from_fw_spec(resume, fw_spec)


        src_dataset = dtoolcore.DataSet.from_uri(source)

        dest_uri = dtoolcore._generate_uri(
            admin_metadata=src_dataset._admin_metadata,
            base_uri=target
        )
        logger.info("Copy from '{}'".format(source))
        logger.info("  to '{}'.".format(dest_uri))

        if not resume:
            # Check if the destination URI is already a dataset
            # and exit gracefully if true.
            if dtoolcore._is_dataset(dest_uri, config_path=None):
                raise FileExistsError(
                    "Dataset already exists: {}".format(dest_uri))

            # If the destination URI is a "file" dataset one needs to check if
            # the path already exists and exit gracefully if true.
            parsed_dataset_uri = dtoolcore.utils.generous_parse_uri(dest_uri)
            if parsed_dataset_uri.scheme == "file":
                if os.path.exists(parsed_dataset_uri.path):
                    raise FileExistsError(
                        "Path already exists: {}".format(parsed_dataset_uri.path))

            logger.info("Resume.")
            copy_func = dtoolcore.copy
        else:
            copy_func = dtoolcore.copy_resume

        target_uri = copy_func(
            src_uri=source,
            dest_base_uri=target,
        )
        logger.info("Copied to '{}'.".format(target_uri))

        target_dataset = dtoolcore.DataSet.from_uri(target_uri)

        return target_dataset


class FetchItemTask(DtoolTask):
    """Fetch a specific item from source dataset to local path.

    This task extends the basic DtoolTask parameters by

    Required params:
        - item_id (str): id of desired item

    Optional params:
        - source (str): URI of source dataset, default: "dataset".
        - dest_dir (str): local destination directory of item. Default: current directory.
        - filename (str)L local filename of item. Default: as returned by dtool

    Fields 'item_id', 'source', 'dest_dir', 'filename' may also be a dict of format
    { 'key': 'some->nested->fw_spec->key' } for looking up value within
    fw_spec instead.
    """

    _fw_name = 'FetchItemTask'
    required_params = [
        *DtoolTask.required_params,
        "item_id"]
    optional_params = [
        *DtoolTask.optional_params,
        "source",
        "dest_dir",
        "filename"]

    def _run_task_internal(self, fw_spec):
        logger = logging.getLogger(__name__)

        item_id = self.get("item_id", None)
        item_id = from_fw_spec(item_id, fw_spec)

        source = self.get("source", "dataset")
        source = from_fw_spec(source, fw_spec)

        dest_dir = self.get("dest_dir", os.path.abspath(os.path.curdir))
        dest_dir = from_fw_spec(dest_dir, fw_spec)

        filename = self.get("filename", None)
        filename = from_fw_spec(filename, fw_spec)

        src_dataset = dtoolcore.DataSet.from_uri(source)
        local_path = src_dataset.item_content_abspath(item_id)

        logger.debug("Local path of '%s' in '%s': %s" % (item_id, source, local_path))

        if filename is None:
            filename = os.path.basename(local_path)
            logger.debug("No explicit filename specified, using '%s'." % filename)

        dest = os.path.join(dest_dir, filename)

        logger.info("Copy from '%s' to '%s'" % (local_path, dest))
        shutil.copy(local_path, dest)

        return src_dataset


# coding: utf-8
#
# storage_tasks.py
#
# Copyright (C) 2020 IMTEK Simulation
# Author: Johannes Hoermann, johannes.hoermann@imtek.uni-freiburg.de
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Tasks that push to and pull from various storage platforms."""

import io
import json
import logging
import os
import pymongo

from ruamel.yaml import YAML
from contextlib import ExitStack

from fireworks.core.firework import FWAction, FiretaskBase
from fireworks.utilities.filepad import FilePad
from fireworks.utilities.dict_mods import get_nested_dict_value, arrow_to_dot

from jlhfw.utils.logging import LoggingContext, _log_nested_dict
from jlhfw.utils.dict import dict_merge

from fireworks.fw_config import FW_LOGGING_FORMAT
from fireworks.utilities.fw_serializers import ENCODING_PARAMS

__author__ = 'Johannes Laurin Hoermann'
__copyright__ = 'Copyright 2020, IMTEK Simulation, University of Freiburg'
__email__ = 'johannes.hoermann@imtek.uni-freiburg.de, johannes.laurin@gmail.com'
__date__ = 'August 27, 2020'

DEFAULT_FORMATTER = logging.Formatter(FW_LOGGING_FORMAT)


def get_fpad(fpad_file):
    if fpad_file:
        return FilePad.from_db_file(fpad_file)
    else:
        return FilePad.auto_load()


class GetObjectFromFilepadTask(FiretaskBase):
    """
    A Firetask to query an object from the filepad and write them to specified
    directory (current working directory if not specified).

    Required params:
        - query (dict or str): mongo db query identifying files to fetch.
          Same as within fireworks.utils.dict_mods, use '->' in dict keys
          for querying nested documents, instead of MongoDB '.' (dot) seperator.
          Do use '.' and NOT '->' within the 'sort_key' field.

    Optional params:
        - sort_key (str): sort key, don't sort per default.
        - sort_direction (int): sort direction, default 'pymongo.DESCENDING'.
            Query can yield multiple objects, only first object in results
            pulled.
        - fizzle_empty_result (bool): fizzle if no file found, default: True
        - filepad_file (str): path to the filepad db config file
        - dest_dir (str): destination directory, default is the current working
            directory.
        - new_file_name (str): if provided, the retrieved file will be
            renamed.
        - metadata_file (str): default: None. Write in YAML-formatted metadata
            of queried object to this file within dest_dir.
        - metadata_fp_source_key (str): if specified, only use nested subset of
            metadata found for filepad object. Default: None or "" uses all.
        - metadata_fw_dest_key (str): forward filepad metadata of queried object
            to this fw_spec field of subsequent Fireworks via FWAction.
            Default: None, metadata not kept in pipeline. Empty string
            points to root.
        - metadata_fw_source_key (str): merge metadata from Filepad object with
            metadata within this fw_spec key before forwarding.
        - fw_supersedes_fp: if True, metadata from metadata_fw_source_key
            supersedes metadata from metadata_fp_source_key. Otherwise, latter
            supersedes former. Default: False
        - metadata_fp_exclusions (dict): mark exclusions from metadata forwarded
            into pipeline with nested dict {'excluded': {'field': True}}.
        - metadata_fw_exclusions (dict): mark exclusions from
            metadata_fw_source_key merged with metadata_fp_source_key.

        - output (str): spec key that will be used to pass output to child
            fireworks. Default: None
        - dict_mod (str, default: '_set'): how to insert output into output
            key, see fireworks.utils.dict_mods
        - propagate (bool, default: None): if True, then set the
            FWAction 'propagate' flag and propagate updated fw_spec not only to
            direct children, but to all descendants down to wokflow's leaves.
        - stored_data (bool, default: False): put outputs into database via
            FWAction.stored_data
        - store_stdlog (bool, default: False): insert log output into database
            (only if 'stored_data' or 'output' is spcified)
        - stdlog_file (str, Default: NameOfTaskClass.log): print log to file
        - loglevel (str, Default: logging.INFO): loglevel for this task
    """
    _fw_name = 'GetObjectsFromFilepadTask'
    required_params = ["query"]
    optional_params = [
        "filepad_file",

        "fizzle_empty_result",
        "sort_direction",
        "sort_key",

        "dest_dir",
        "new_file_name",

        "metadata_file",

        "metadata_fp_source_key",
        "metadata_fw_source_key",
        "metadata_fw_dest_key",
        "fw_supersedes_fp",
        "metadata_fp_exclusions",
        "metadata_fw_exclusions",

        "stored_data",
        "output",
        "dict_mod",
        "propagate",
        "stdlog_file",
        "store_stdlog",
        "loglevel"
        ]

    def run_task(self, fw_spec):
        fpad = get_fpad(self.get("filepad_file", None))
        dest_dir = self.get("dest_dir", os.path.abspath(os.path.curdir))
        new_file_name = self.get("new_file_name", None)
        query = self.get("query", {})
        sort_key = self.get("sort_key", None)
        sort_direction = self.get("sort_direction", pymongo.DESCENDING)
        fizzle_empty_result = self.get("fizzle_empty_result", True)
        metadata_file = self.get("metadata_file", None)

        metadata_fp_source_key = self.get("metadata_fp_source_key", None)
        metadata_fw_source_key = self.get("metadata_fw_source_key", None)
        metadata_fw_dest_key = self.get("metadata_fw_dest_key", None)
        fw_supersedes_fp = self.get("fw_supersedes_fp", False)
        metadata_fp_exclusions = self.get("metadata_fp_exclusions", {})
        metadata_fw_exclusions = self.get("metadata_fw_exclusions", {})

        # generic parameters
        stored_data = self.get('stored_data', False)
        output_key = self.get('output', None)
        dict_mod = self.get('dict_mod', '_set')
        propagate = self.get('propagate', False)

        stdlog_file = self.get('stdlog_file', '{}.log'.format(self._fw_name))
        store_stdlog = self.get('store_stdlog', False)
        loglevel = self.get('loglevel', logging.INFO)

        mod_spec = []
        with ExitStack() as stack:

            if store_stdlog:
                stdlog_stream = io.StringIO()
                logh = logging.StreamHandler(stdlog_stream)
                logh.setFormatter(DEFAULT_FORMATTER)
                stack.enter_context(
                    LoggingContext(handler=logh, level=loglevel, close=False))

            # logging to dedicated log file if desired
            if stdlog_file:
                logfh = logging.FileHandler(
                    stdlog_file, mode='a', **ENCODING_PARAMS)
                logfh.setFormatter(DEFAULT_FORMATTER)
                stack.enter_context(
                    LoggingContext(handler=logfh, level=loglevel, close=True))

            logger = logging.getLogger(__name__)

            if isinstance(query, dict):
                query = arrow_to_dot(query)
            # else assumed str

            if sort_key is None:
                cursor = fpad.filepad.find(query)
            else:
                cursor = fpad.filepad.find(query).sort(sort_key, sort_direction)

            try:
                d = next(cursor)
            except StopIteration:
                if fizzle_empty_result:
                    raise ValueError("Query yielded empty result! (query: {:s})".format(
                        json.dumps(query)))
                else:
                    logger.info("Query yielded empty result! (query: {:s})".format(json.dumps(query)))
            else:
                file_contents, doc = fpad._get_file_contents(d)
                logger.debug("Metadata of queried object:")
                _log_nested_dict(logger.debug, doc)

                file_name = new_file_name if new_file_name else doc["original_file_name"]
                with open(os.path.join(dest_dir, file_name), "wb") as f:
                    f.write(file_contents)

                if metadata_file:
                    metadata_file_path = os.path.join(dest_dir, metadata_file)
                    logger.debug("Writing metadata of queried object to '{}'.".format(metadata_file_path))
                    with open(metadata_file_path, "w") as f:
                        yaml = YAML()
                        yaml.dump(doc, f)

                if metadata_fw_dest_key:
                    if metadata_fp_source_key and metadata_fp_source_key != "":
                        fp_metadata = get_nested_dict_value(doc, metadata_fp_source_key)
                        logger.debug(
                            "Forwarding '{}' subset of filepad object metadata:".format(metadata_fp_source_key))
                        _log_nested_dict(logger.debug, fp_metadata)
                    else:
                        fp_metadata = doc
                        logger.debug("Forwarding filepad object metadata.")

                    if metadata_fw_source_key and metadata_fw_source_key == "":
                        fw_metadata = fw_spec
                        logger.debug("Merging with fw_spec.")
                    elif metadata_fw_source_key:
                        fw_metadata = get_nested_dict_value(fw_spec, metadata_fw_source_key)
                        logger.debug("Merging with '{}' subset of fw_spec:".format(metadata_fw_source_key))
                        _log_nested_dict(logger.debug, fw_metadata)
                    else:
                        fw_metadata = {}

                    if fw_supersedes_fp:
                        metadata = dict_merge(fp_metadata, fw_metadata,
                                              exclusions=metadata_fp_exclusions,
                                              merge_exclusions=metadata_fw_exclusions)
                    else:
                        metadata = dict_merge(fw_metadata, fp_metadata,
                                              exclusions=metadata_fw_exclusions,
                                              merge_exclusions=metadata_fp_exclusions)

                    logger.debug("Forwarding merge into '{}':".format(metadata_fw_dest_key))
                    _log_nested_dict(logger.debug, metadata)

                    mod_spec.append({dict_mod: {metadata_fw_dest_key: metadata}})

        # end of ExitStack context
        output = {}
        if store_stdlog:
            stdlog_stream.flush()
            output['stdlog'] = stdlog_stream.getvalue()

        fw_action = FWAction()

        if stored_data:
            fw_action.stored_data = output

        if hasattr(fw_action, 'propagate') and propagate:
            fw_action.propagate = propagate

        if output_key:  # inject into fw_spec
            mod_spec.append({dict_mod: {output_key: output}})

        if len(mod_spec) > 0:
            fw_action.mod_spec = mod_spec

        return fw_action
